@@ -233,7 +233,7 @@ public:
   u32 nrels = 0;
   u32 unwind_offset = 0;
   u32 nunwind = 0;
-  Subsection<E> *replacer; // Used if is_coalesced is true
+  Subsection<E> *replacer = nullptr; // Used if is_coalesced is true
 
   std::atomic_uint8_t p2align = 0;
   std::atomic_bool is_alive = true;
