@@ -72,7 +72,8 @@ inline std::string mold_version;
 extern std::string mold_version_string;
 extern std::string mold_git_hash;
 
-std::string_view errno_string();
+std::string errno_string();
+std::string get_self_path();
 void cleanup();
 void install_signal_handler();
 i64 get_default_thread_count();
@@ -219,16 +220,29 @@ inline i64 sign_extend(u64 val, i64 size) {
 
 template <typename T, typename Compare = std::less<T>>
 void update_minimum(std::atomic<T> &atomic, u64 new_val, Compare cmp = {}) {
-  T old_val = atomic;
+  T old_val = atomic.load(std::memory_order_relaxed);
   while (cmp(new_val, old_val) &&
-         !atomic.compare_exchange_weak(old_val, new_val));
+         !atomic.compare_exchange_weak(old_val, new_val,
+                                       std::memory_order_relaxed));
 }
 
 template <typename T, typename Compare = std::less<T>>
 void update_maximum(std::atomic<T> &atomic, u64 new_val, Compare cmp = {}) {
-  T old_val = atomic;
+  T old_val = atomic.load(std::memory_order_relaxed);
   while (cmp(old_val, new_val) &&
-         !atomic.compare_exchange_weak(old_val, new_val));
+         !atomic.compare_exchange_weak(old_val, new_val,
+                                       std::memory_order_relaxed));
+}
+
+// An optimized "mark" operation for parallel mark-and-sweep algorithms.
+// Returns true if `visited` was false and updated to true.
+inline bool fast_mark(std::atomic<bool> &visited) {
+  // A relaxed load + branch (assuming miss) takes only around 20 cycles,
+  // while an atomic RMW can easily take hundreds on x86. We note that it's
+  // common that another thread beat us in marking, so doing an optimistic
+  // early test tends to improve performance in the ~20% ballpark.
+  return !visited.load(std::memory_order_relaxed) &&
+         !visited.exchange(true, std::memory_order_relaxed);
 }
 
 template <typename T, typename U>
@@ -553,7 +567,6 @@ private:
   std::vector<std::string> strings;
   std::unique_ptr<TrieNode> root;
   std::vector<std::pair<Glob, u32>> globs;
-  std::vector<u32> values;
   std::once_flag once;
   bool is_compiled = false;
 };
@@ -731,7 +744,8 @@ public:
   static MappedFile *open(C &ctx, std::string path);
   static MappedFile *must_open(C &ctx, std::string path);
 
-  ~MappedFile();
+  ~MappedFile() { unmap(); }
+  void unmap();
 
   MappedFile *slice(C &ctx, std::string name, u64 start, u64 size);
 
@@ -778,7 +792,6 @@ MappedFile<C> *MappedFile<C>::open(C &ctx, std::string path) {
   if (path.starts_with('/') && !ctx.arg.chroot.empty())
     path = ctx.arg.chroot + "/" + path_clean(path);
 
-
   i64 fd;
 #ifdef _WIN32
     fd = ::_open(path.c_str(), O_RDONLY);
@@ -851,8 +864,8 @@ MappedFile<C>::slice(C &ctx, std::string name, u64 start, u64 size) {
 }
 
 template <typename C>
-MappedFile<C>::~MappedFile() {
-  if (size == 0 || parent)
+void MappedFile<C>::unmap() {
+  if (size == 0 || parent || !data)
     return;
 
 #ifdef _WIN32
@@ -862,6 +875,8 @@ MappedFile<C>::~MappedFile() {
 #else
   munmap(data, size);
 #endif
+
+  data = nullptr;
 }
 
 } // namespace mold
