@@ -1,18 +1,7 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
-command -v dwarfdump >& /dev/null || { echo skipped; exit; }
+command -v dwarfdump >& /dev/null || skip
 
 cat <<EOF | $CXX -c -o $t/a.o -g -xc++ -
 #include <iostream>
@@ -38,5 +27,3 @@ $CXX -B. -o $t/exe $t/a.o $t/b.o $t/c.o -g
 $QEMU $t/exe | grep -q 'Hello world'
 
 dwarfdump $t/exe > /dev/null
-
-echo OK
