@@ -1,25 +1,12 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
-[ $MACHINE = $(uname -m) ] || { echo skipped; exit; }
-[ $MACHINE = riscv64 -o $MACHINE = riscv32 ] && { echo skipped; exit; }
-[ $MACHINE = sparc64 ] && { echo skipped; exit; }
+[ $MACHINE = $HOST ] || skip
+[ $MACHINE = riscv64 -o $MACHINE = riscv32 -o $MACHINE = sparc64 ] && skip
 
-command -v gdb >& /dev/null || { echo skipped; exit; }
+command -v gdb >& /dev/null || skip
 
-echo 'int main() {}' | $CC -o /dev/null -xc -gdwarf-2 -g - >& /dev/null ||
-  { echo skipped; exit; }
+echo 'int main() {}' | $CC -o /dev/null -xc -gdwarf-2 -g - >& /dev/null || skip
 
 cat <<EOF | $CC -c -o $t/a.o -fPIC -g -ggnu-pubnames -gdwarf-2 -xc - -ffunction-sections
 void hello2();
@@ -67,5 +54,3 @@ grep -q 'hello2 () at .*<stdin>:7' $t/log
 grep -q 'hello () at .*<stdin>:4' $t/log
 grep -q 'greet () at .*<stdin>:8' $t/log
 grep -q 'main () at .*<stdin>:4' $t/log
-
-echo OK
