@@ -1,18 +1,8 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
-[ $MACHINE = ppc64le ] && { echo skipped; exit; }
+[ $MACHINE = ppc64 ] && skip
+[ $MACHINE = ppc64le ] && skip
 
 cat <<EOF | $CC -shared -o $t/a.so -xc -
 int foo = 3;
@@ -37,5 +27,3 @@ $QEMU $t/exe | grep -q '3 5'
 ! $CC -B. -o $t/exe $t/a.so $t/b.o -no-pie -Wl,-z,nocopyreloc 2> $t/log || false
 
 grep -q 'recompile with -fPIC' $t/log
-
-echo OK
