@@ -1,24 +1,13 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 # Skip if 32 bits as we use very large addresses in this test.
-[ $MACHINE = i386 -o $MACHINE = i686 ] && { echo skipped; exit; }
-[ $MACHINE = arm32 ] && { echo skipped; exit; }
-[ $MACHINE = riscv32 ] && { echo skipped; exit; }
+[ $MACHINE = i386 ] && skip
+[ $MACHINE = arm32 ] && skip
+[ $MACHINE = riscv32 ] && skip
 
 # It looks like SPARC's runtime can't handle PLT if it's too far from GOT.
-[ $MACHINE = sparc64 ] && { echo skipped; exit; }
+[ $MACHINE = sparc64 ] && skip
 
 cat <<EOF > $t/a.c
 #include <stdio.h>
@@ -49,15 +38,13 @@ EOF
 $CC -c -o $t/c.o $t/a.c -O0
 $CC -c -o $t/d.o $t/b.c -O0
 
-$CC -B. -o $t/exe $t/c.o $t/d.o \
+$CC -B. -o $t/exe1 $t/c.o $t/d.o \
   -Wl,--section-start=.low=0x10000000,--section-start=.high=0x20000000
-$QEMU $t/exe | grep -q 'main fn1 fn3 fn2 fn4'
+$QEMU $t/exe1 | grep -q 'main fn1 fn3 fn2 fn4'
 
 $CC -c -o $t/e.o $t/a.c -O2
 $CC -c -o $t/f.o $t/b.c -O2
 
-$CC -B. -o $t/exe $t/e.o $t/f.o \
+$CC -B. -o $t/exe2 $t/e.o $t/f.o \
   -Wl,--section-start=.low=0x10000000,--section-start=.high=0x20000000
-$QEMU $t/exe | grep -q 'main fn1 fn3 fn2 fn4'
-
-echo OK
+$QEMU $t/exe2 | grep -q 'main fn1 fn3 fn2 fn4'
