@@ -1,16 +1,5 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 cat <<EOF | $CC -fPIC -c -o $t/a.o -xc -
 void foo() {}
@@ -41,5 +30,3 @@ EOF
 $CC -B. -o $t/exe3 $t/a.o $t/b.so -Wl,--version-script=$t/d.ver
 nm -g $t/exe3 > $t/log3
 ! grep -q foo $t/log3 || false
-
-echo OK
